#!/usr/bin/env pike

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * This is gnugoclient, a program to connect go engines to go        *
 * servers. See http://www.lysator.liu.se/~gunnar/gnugoclient        *
 * for information.                                                  *
 *                                                                   *
 * Copyright 1998-2004 by Gunnar Farnebck.                          *
 * E-mail: gunnar@lysator.liu.se                                     *
 *                                                                   *
 * This program is free software; you can redistribute it and/or     *
 * modify it under the terms of the GNU General Public License as    *
 * published by the Free Software Foundation - version 2             *
 *                                                                   *
 * This program is distributed in the hope that it will be useful,   *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of    *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the     *
 * GNU General Public License in file COPYING for more details.      *
 *                                                                   *
 * You should have received a copy of the GNU General Public         *
 * License along with this program; if not, write to the Free        *
 * Software Foundation, Inc., 59 Temple Place - Suite 330,           *
 * Boston, MA 02111, USA.                                            *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

inherit "GtpClient.pike";

void game_finished()
{
    exit(0);
}

// Change some default option settings before parsing the command line.
void get_options(array(string) args)
{
    options->server_address = "nngs.computer-go.jp";
    options->server_name = "Gifu";
    options->sgf_save_time_info = "on",

    ::get_options(args);
}

int acceptable_time(int main_time, int byo_yomi_time)
{
    return main_time == 40 && byo_yomi_time == 0;
}

string why_unacceptable_time(int main_time, int byo_yomi_time)
{
    return "The Gifu tournament rules require 40/0 time settings.";
}

void start_game()
{
    tell_server("komi 6.5");
    ::start_game();
}

void gifu_score_game()
{
    array(string) messages = ({});
    if (has_value(known_commands, "final_score"))
    {
	string score = engine->send_command_sync("final_score");
	messages += ({"final_score: " + score});
    }
    
    if (has_value(known_commands, "final_status_list"))
    {
	string dead = engine->send_command_sync("final_status_list dead");
	array(string) dead_groups = process_stone_list(dead, "dead");
	if (sizeof(dead_groups) == 0)
	    messages += ({"All stones seem to be alive."});
	else if (sizeof(dead_groups) == 1)
	    messages += ({"The " + dead_groups[0] + " string is dead."});
	else
	    messages += ({"The " + dead_groups * ", " + " strings are dead."});
    
	string seki = engine->send_command_sync("final_status_list seki");
	array(string) seki_groups = process_stone_list(seki, "seki");
	if (sizeof(seki_groups) > 0)
	    tell_opponent(seki_groups * ", " + " live in seki.");
    }

    write("*************************\r\n");
    
    foreach (messages, string s)
    {
	add_sgf_comment(s);
	write(s + "\r\n");
    }

    write("*************************\r\n");
    write("Game finished, exiting.\r\n");
}

void parse_server_message_old(string s)
{
    string komistring;
    string declinestring;
    if (has_prefix(s, "9 You can check your score"))
    {
	gifu_score_game();
	engine->shutdown();
	tell_server("quit");
	sleep(5);
	exit(0);
    }
    // Opponent wants to change komi. Only accept 6.5.
    else if (sscanf(s, "9 Use \"komi %s\" to accept, or \"%s\"",
		    komistring, declinestring) == 2)
    {
	float komi = (float) komistring;
	if (komi == 6.5)
	{
	    tell_server("komi " + komi);
	    if (engine)
	    {
		engine->send_command("komi " + komistring);
		add_root_property("KM", komistring);
	    }
	}
	else
	{
	    tell_server(declinestring);
	}
    }
    else
	::parse_server_message_old(s);
}

// No stored games interference when playing the tournament.
void update_stored_games()
{
    return;
}
